@@ -1,6 +1,6 @@
-from asyncio import TimeoutError
+from asyncio import TimeoutError, get_running_loop
 from time import time
-from typing import List, Optional, Union
+from typing import List, Optional, Union, Callable
 from random import randint
 
 from discord import Embed, Emoji, Member, PartialEmoji, Role, TextChannel, User
@@ -28,13 +28,19 @@ from .errors import (
 class TimedOut:
     def __init__(
         self,
-        ctx,
-        buttonContext,
-        timeTaken,
-        lastContent,
-        lastEmbed,
-        successfulUsers,
-        failedUsers,
+        ctx: Union[
+            InteractionContext,
+            Context,
+            TextChannel,
+            User,
+            Member,
+        ],
+        buttonContext: ComponentContext,
+        timeTaken: int,
+        lastContent: str,
+        lastEmbed: Embed,
+        successfulUsers: List[User],
+        failedUsers: List[User],
     ):
         self.ctx = ctx
         self.buttonContext = buttonContext
@@ -68,6 +74,7 @@ class Paginator:
             ]
         ] = None,
         dm: Optional[bool] = False,
+        customActionRow: Optional[List[Union[dict, Callable]]] = None,
         timeout: Optional[int] = None,
         disableAfterTimeout: Optional[bool] = True,
         deleteAfterTimeout: Optional[bool] = False,
@@ -75,8 +82,10 @@ class Paginator:
         useButtons: Optional[bool] = True,
         useIndexButton: Optional[bool] = None,
         useLinkButton: Optional[bool] = False,
+        useQuitButton: Optional[bool] = False,
         useFirstLast: Optional[bool] = True,
         useOverflow: Optional[bool] = True,
+        useNotYours: Optional[bool] = True,
         firstLabel: Optional[str] = "",
         prevLabel: Optional[str] = "",
         indexLabel: Optional[str] = "Page",
@@ -85,17 +94,20 @@ class Paginator:
         linkLabel: Optional[Union[str, List[str]]] = "",
         linkURL: Optional[Union[str, List[str]]] = "",
         customButtonLabel: Optional[str] = None,
+        quitButtonLabel: Optional[str] = "Quit",
         firstEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = "⏮️",
         prevEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = "◀",
         nextEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = "▶",
         lastEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = "⏭️",
         customButtonEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = None,
+        quitButtonEmoji: Optional[Union[Emoji, PartialEmoji, dict, str]] = None,
         firstStyle: Optional[Union[ButtonStyle, int]] = 1,
         prevStyle: Optional[Union[ButtonStyle, int]] = 1,
         indexStyle: Optional[Union[ButtonStyle, int]] = 2,
         nextStyle: Optional[Union[ButtonStyle, int]] = 1,
         lastStyle: Optional[Union[ButtonStyle, int]] = 1,
         customButtonStyle: Optional[Union[ButtonStyle, int]] = 2,
+        quitButtonStyle: Optional[Union[ButtonStyle, int]] = 4,
     ) -> TimedOut:
         self.bot = bot
         self.ctx = ctx
@@ -105,7 +117,6 @@ class Paginator:
         self.authorOnly = authorOnly
         self.onlyFor = onlyFor
         self.dm = dm
-        self.useOverflow = useOverflow
         self.timeout = timeout
         self.disableAfterTimeout = disableAfterTimeout
         self.deleteAfterTimeout = deleteAfterTimeout
@@ -113,14 +124,17 @@ class Paginator:
         self.useButtons = useButtons
         self.useIndexButton = useIndexButton
         self.useLinkButton = useLinkButton
+        self.useQuitButton = useQuitButton
         self.useFirstLast = useFirstLast
+        self.useOverflow = useOverflow
+        self.useNotYours = useNotYours
         self.labels = [firstLabel, prevLabel, indexLabel, nextLabel, lastLabel]
         self.links = [linkLabel, linkURL]
-        self.customButtonLabel = customButtonLabel
+        self.custom = [customButtonStyle, customButtonLabel, customButtonEmoji]
+        self.quit = [quitButtonStyle, quitButtonLabel, quitButtonEmoji]
         self.emojis = [firstEmoji, prevEmoji, nextEmoji, lastEmoji]
-        self.customButtonEmoji = customButtonEmoji
         self.styles = [firstStyle, prevStyle, indexStyle, nextStyle, lastStyle]
-        self.customButtonStyle = customButtonStyle
+        self.customActionRow = customActionRow
 
         self.top = len(self.pages)  # limit of the paginator
         self.multiContent = False
@@ -133,6 +147,8 @@ class Paginator:
             self.successfulUsers = [ctx]
         self.failedUsers = []
 
+        self.usedLink = self.usedCustom = self.usedQuit = False
+
         self.index = 1
         self.random_id = randint(1, 999999)
 
@@ -162,13 +178,17 @@ class Paginator:
             "discord.User/Role, or list(discord.User/Role)",
         )
         self.incdata("dm", self.dm, bool, "bool")
-        self.incdata("useOverflow", self.useOverflow, bool, "bool")
         self.incdata("timeout", self.timeout, (int, type(None)), "int")
         bools = {
             "disableAfterTimeout": self.disableAfterTimeout,
             "deleteAfterTimeout": self.deleteAfterTimeout,
             "useSelect": self.useSelect,
+            "useButtons": self.useButtons,
             "useLinkButton": self.useLinkButton,
+            "useQuitButton": self.useQuitButton,
+            "useFirstLast": self.useFirstLast,
+            "useOverflow": self.useOverflow,
+            "useNotYours": self.useNotYours,
         }
         for b in bools:
             self.incdata(
@@ -186,6 +206,7 @@ class Paginator:
             "lastLabel": lastLabel,
             "linkLabel": linkLabel,
             "linkURL": linkURL,
+            "quitButtonLabel": quitButtonLabel,
         }
         for label in labels:
             self.incdata(
@@ -194,8 +215,12 @@ class Paginator:
                 str,
                 "str",
             )
+        self.incdata("customButtonLabel", customButtonLabel, (str, type(None)), "str")
         self.incdata(
-            "customButtonLabel", self.customButtonLabel, (str, type(None)), "str"
+            "quitButtonEmoji",
+            quitButtonEmoji,
+            (Emoji, PartialEmoji, dict, str, type(None)),
+            "Emoji, PartialEmoji, dict, str, or None",
         )
         emojis = {
             "firstEmoji": firstEmoji,
@@ -216,6 +241,7 @@ class Paginator:
             "indexStyle": indexStyle,
             "nextStyle": nextStyle,
             "lastStyle": lastStyle,
+            "quitButtonStyle": quitButtonStyle,
         }
         for style in styles:
             self.incdata(
@@ -224,6 +250,13 @@ class Paginator:
                 (ButtonStyle, int),
                 "ButtonStyle or int",
             )
+        self.incdata(
+            "customActionRow",
+            self.customActionRow,
+            (list, type(None)),
+            "a list with the action row, then the function",
+        )
+
         if self.useIndexButton and not self.useButtons:
             BadButtons("Index button cannot be used with useButtons=False!")
             self.useIndexButton = False
@@ -232,7 +265,7 @@ class Paginator:
             BadOnly()
             self.authorOnly = False
 
-        if self.customButtonLabel is not None:
+        if customButtonLabel is not None:
             self.useCustomButton = True
 
         if self.useSelect and len(self.pages) > 25:
@@ -241,6 +274,7 @@ class Paginator:
                 self.useIndexButton = True
 
     async def run(self) -> TimedOut:
+        self.usedLink = self.usedCustom = self.usedQuit = False
         try:
             if self.dm:
                 if isinstance(self.ctx, InteractionContext) or isinstance(
@@ -249,16 +283,7 @@ class Paginator:
                     msg = await self.ctx.author.send(
                         content=self.content[0] if self.multiContent else self.content,
                         embed=self.pages[0],
-                        components=list(
-                            filter(
-                                None,
-                                [
-                                    self.select_row() if self.useSelect else None,
-                                    self.buttons_row() if self.useButtons else None,
-                                    self.overflow_row() if self.useOverflow else None,
-                                ],
-                            )
-                        ),
+                        components=self.components(),
                         hidden=self.hidden,
                     )
                     await self.ctx.send("Check your DMs!", hidden=True)
@@ -266,16 +291,7 @@ class Paginator:
                     msg = await self.ctx.send(
                         content=self.content[0] if self.multiContent else self.content,
                         embed=self.pages[0],
-                        components=list(
-                            filter(
-                                None,
-                                [
-                                    self.select_row() if self.useSelect else None,
-                                    self.buttons_row() if self.useButtons else None,
-                                    self.overflow_row() if self.useOverflow else None,
-                                ],
-                            )
-                        ),
+                        components=self.components(),
                         hidden=self.hidden,
                     )
                 else:
@@ -288,16 +304,7 @@ class Paginator:
                 msg = await self.ctx.send(
                     content=self.content[0] if self.multiContent else self.content,
                     embed=self.pages[0],
-                    components=list(
-                        filter(
-                            None,
-                            [
-                                self.select_row() if self.useSelect else None,
-                                self.buttons_row() if self.useButtons else None,
-                                self.overflow_row() if self.useOverflow else None,
-                            ],
-                        )
-                    ),
+                    components=self.components(),
                     hidden=self.hidden,
                 )
         except TypeError:
@@ -308,32 +315,14 @@ class Paginator:
                     msg = await self.ctx.author.send(
                         content=self.content[0] if self.multiContent else self.content,
                         embed=self.pages[0],
-                        components=list(
-                            filter(
-                                None,
-                                [
-                                    self.select_row() if self.useSelect else None,
-                                    self.buttons_row() if self.useButtons else None,
-                                    self.overflow_row() if self.useOverflow else None,
-                                ],
-                            )
-                        ),
+                        components=self.components(),
                     )
                     await self.ctx.send("Check your DMs!", hidden=True)
                 elif isinstance(self.ctx, userUser):
                     msg = await self.ctx.send(
                         content=self.content[0] if self.multiContent else self.content,
                         embed=self.pages[0],
-                        components=list(
-                            filter(
-                                None,
-                                [
-                                    self.select_row() if self.useSelect else None,
-                                    self.buttons_row() if self.useButtons else None,
-                                    self.overflow_row() if self.useOverflow else None,
-                                ],
-                            )
-                        ),
+                        components=self.components(),
                     )
                 else:
                     IncorrectDataType(
@@ -343,19 +332,9 @@ class Paginator:
                 msg = await self.ctx.send(
                     content=self.content[0] if self.multiContent else self.content,
                     embed=self.pages[0],
-                    components=list(
-                        filter(
-                            None,
-                            [
-                                self.select_row() if self.useSelect else None,
-                                self.buttons_row() if self.useButtons else None,
-                                self.overflow_row() if self.useOverflow else None,
-                            ],
-                        )
-                    ),
+                    components=self.components(),
                 )
-        # handling the interaction
-        tmt = True  # stop listening when timeout expires
+        tmt = True
         start = time()
         buttonContext = None
         while tmt:
@@ -363,16 +342,7 @@ class Paginator:
                 buttonContext: ComponentContext = await wait_for_component(
                     self.bot,
                     check=self.check,
-                    components=list(
-                        filter(
-                            None,
-                            [
-                                self.select_row() if self.useSelect else None,
-                                self.buttons_row() if self.useButtons else None,
-                                self.overflow_row() if self.useOverflow else None,
-                            ],
-                        )
-                    ),
+                    components=self.components(),
                     timeout=self.timeout,
                 )
                 if buttonContext.author not in self.successfulUsers:
@@ -387,22 +357,43 @@ class Paginator:
                     self.index = self.top
                 elif buttonContext.custom_id == f"select{self.random_id}":
                     self.index = int(buttonContext.selected_options[0])
+                elif buttonContext.custom_id == f"quit{self.random_id}":
+                    tmt = False
+                    end = time()
+                    if self.deleteAfterTimeout and not self.hidden:
+                        await buttonContext.origin_message.delete()
+                    elif self.disableAfterTimeout and not self.hidden:
+                        components = self.components()
+                        for row in components:
+                            for component in row["components"]:
+                                component["disabled"] = True
+                        await buttonContext.edit_origin(components=components)
+                    timeTaken = round(end - start)
+                    lastContent = (
+                        self.content
+                        if isinstance(self.content, (str, type(None)))
+                        else self.content[self.index - 1]
+                    )
+                    lastEmbed = self.pages[self.index - 1]
+                    return TimedOut(
+                        self.ctx,
+                        buttonContext,
+                        timeTaken,
+                        lastContent,
+                        lastEmbed,
+                        self.successfulUsers,
+                        self.failedUsers,
+                    )
+                elif self.customActionRow is not None:
+                    await self.customActionRow[1](self, buttonContext)
+                    continue
 
                 await buttonContext.edit_origin(
                     content=self.content[self.index - 1]
                     if self.multiContent
                     else self.content,
                     embed=self.pages[self.index - 1],
-                    components=list(
-                        filter(
-                            None,
-                            [
-                                self.select_row() if self.useSelect else None,
-                                self.buttons_row() if self.useButtons else None,
-                                self.overflow_row() if self.useOverflow else None,
-                            ],
-                        )
-                    ),
+                    components=self.components(),
                 )
             except TimeoutError:
                 tmt = False
@@ -410,16 +401,7 @@ class Paginator:
                 if self.deleteAfterTimeout and not self.hidden:
                     await msg.edit(components=None)
                 elif self.disableAfterTimeout and not self.hidden:
-                    components = components = list(
-                        filter(
-                            None,
-                            [
-                                self.select_row() if self.useSelect else None,
-                                self.buttons_row() if self.useButtons else None,
-                                self.overflow_row() if self.useOverflow else None,
-                            ],
-                        )
-                    )
+                    components = self.components()
                     for row in components:
                         for component in row["components"]:
                             component["disabled"] = True
@@ -444,6 +426,13 @@ class Paginator:
     def check(self, buttonContext) -> bool:
         if self.authorOnly and buttonContext.author.id != self.ctx.author.id:
             if buttonContext.author not in self.failedUsers:
+                if self.useNotYours:
+                    get_running_loop().create_task(
+                        buttonContext.send(
+                            f"{buttonContext.author.mention}, this paginator is not for you!",
+                            hidden=True,
+                        )
+                    )
                 self.failedUsers.append(buttonContext.author)
             return False
         if self.onlyFor is not None:
@@ -460,6 +449,13 @@ class Paginator:
                     check = check or self.onlyFor in buttonContext.author.roles
 
             if not check:
+                if self.useNotYours:
+                    get_running_loop().create_task(
+                        buttonContext.send(
+                            f"{buttonContext.author.mention}, this paginator is not for you!",
+                            hidden=True,
+                        )
+                    )
                 self.failedUsers.append(buttonContext.author)
                 return False
 
@@ -528,6 +524,7 @@ class Paginator:
         if not self.useIndexButton:
             controlButtons.pop(1)
         if self.useFirstLast:
+            # First button
             controlButtons.insert(
                 0,
                 create_button(
@@ -538,6 +535,7 @@ class Paginator:
                     emoji=self.emojis[0],
                 ),
             )
+            # Last button
             controlButtons.append(
                 create_button(
                     style=self.styles[4],
@@ -555,24 +553,38 @@ class Paginator:
             )
             if len(controlButtons) < 5:
                 controlButtons.append(linkButton)
+                self.usedLink = True
             elif not self.useOverflow:
                 raise TooManyButtons()
         if self.useCustomButton:
             customButton = create_button(
-                style=self.customButtonStyle,
-                label=self.customButtonLabel,
+                style=self.custom[0],
+                label=self.custom[1],
                 disabled=True,
-                emoji=self.customButtonEmoji,
+                emoji=self.custom[2],
             )
             if len(controlButtons) < 5:
                 controlButtons.append(customButton)
+                self.usedCustom = True
+            elif not self.useOverflow:
+                raise TooManyButtons()
+        if self.useQuitButton:
+            quitButton = create_button(
+                style=self.quit[0],
+                label=self.quit[1],
+                emoji=self.quit[2],
+                custom_id=f"quit{self.random_id}",
+            )
+            if len(controlButtons) < 5:
+                controlButtons.append(quitButton)
+                self.usedCustom = True
             elif not self.useOverflow:
                 raise TooManyButtons()
         return create_actionrow(*controlButtons)
 
     def overflow_row(self) -> dict:
         controlButtons = []
-        if self.useLinkButton:
+        if self.useLinkButton and not self.usedLink:
             linkButton = create_button(
                 style=5,
                 label=self.links[0][0] if self.multiLabel else self.links[0],
@@ -580,17 +592,39 @@ class Paginator:
             )
             if len(controlButtons) < 5:
                 controlButtons.append(linkButton)
-        if self.useCustomButton:
+        if self.useCustomButton and not self.usedCustom:
             customButton = create_button(
-                style=self.customButtonStyle,
-                label=self.customButtonLabel,
+                style=self.custom[0],
+                label=self.custom[1],
                 disabled=True,
-                emoji=self.customButtonEmoji,
+                emoji=self.custom[2],
             )
             if len(controlButtons) < 5:
                 controlButtons.append(customButton)
+        if self.useQuitButton and not self.usedCustom:
+            quitButton = create_button(
+                style=self.quit[0],
+                label=self.quit[1],
+                emoji=self.quit[2],
+                custom_id=f"quit{self.random_id}",
+            )
+            if len(controlButtons) < 5:
+                controlButtons.append(quitButton)
         return None if controlButtons == [] else create_actionrow(*controlButtons)
 
+    def components(self) -> list:
+        return list(
+            filter(
+                None,
+                [
+                    self.select_row() if self.useSelect else None,
+                    self.buttons_row() if self.useButtons else None,
+                    self.overflow_row() if self.useOverflow else None,
+                    None if self.customActionRow is None else self.customActionRow[0],
+                ],
+            )
+        )
+
     def incdata(self, name, arg, sup, supstr) -> None:
         if isinstance(arg, tuple) and isinstance(sup, tuple):
             for a in arg:
